//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/XML/Serializer.h
//! @brief     Defines class Serializer
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_XML_SERIALIZER_H
#define BORNAGAIN_GUI_MODEL_XML_SERIALIZER_H

#include <QXmlStreamWriter> // used in every including file
#include "GUI/Model/Descriptor/SelectionProperty.h"
#include <functional>
#include <heinz/Vectors3D.h>

class DoubleProperty;
class VectorProperty;
class UIntProperty;
class AxisProperty;

// #bamigration docu
class Serializer {
public:
    Serializer(QXmlStreamWriter* w)
        : m_w(w)
    {
    }
    Serializer(QXmlStreamReader* r)
        : m_r(r)
    {
    }

    // #bamigration docu
    //! No meaning for reading
    void writeVersion(unsigned version) const;

    //! No meaning for writing
    void assertVersion(unsigned expectedVersion) const;
    bool isReading() const;
    bool isWriting() const;

    void rw(const QString& tag, QString& val);
    void rw(const QString& tag, QColor& col);
    void rw(const QString& tag, bool& val);
    void rw(const QString& tag, int& val);
    void rw(const QString& tag, double& val);
    void rw(const QString& tag, R3& val);
    void rw(const QString& tag, QByteArray& val);
    void rw(DoubleProperty& d);
    void rw(VectorProperty& d);
    void rw(UIntProperty& d);
    void rw(const QString& tag, AxisProperty& d);

    //! serialize a list with a known and fixed type which is not a base class, but the real class
    template <typename T, typename... Args>
    void rw(const QString& tag, QVector<T>& vec, Args... argsForConstructor)
    {
        if (m_w) {
            m_w->writeStartElement(tag);
            m_w->writeAttribute("n", QString::number(vec.size()));
            int i = 0;
            for (auto* p : vec) {
                m_w->writeStartElement("E" + QString::number(i++));
                p->serialize(*this);
                m_w->writeEndElement();
            }
            m_w->writeEndElement();
        } else {
            m_r->readNextStartElement();
            assertCurrentTag(tag);
            const int n = m_r->attributes().value("n").toInt();
            for (int i = 0; i < n; i++) {
                goToStartElementOfTag("E" + QString::number(i));
                vec << new typename std::remove_pointer<T>::type(argsForConstructor...);
                vec.last()->serialize(*this);
                gotoEndElementOfTag("E" + QString::number(i));
            }
            gotoEndElementOfTag(tag);
        }
    }

    //! serialize a list with a base class type, which can contain derived types. For this we need
    //! the catalog of possible classes/types to identify the current type and to re-create this
    //! type after serialization
    template <typename Catalog, typename... Args>
    void rw(const QString& tag, QVector<typename Catalog::CatalogedType*>& vec,
            Args... argsForConstructor)
    {
        if (m_w) {
            m_w->writeStartElement(tag);
            m_w->writeAttribute("n", QString::number(vec.size()));
            int i = 0;
            for (auto* p : vec) {
                write<Catalog>("E" + QString::number(i++), p);
            }
            m_w->writeEndElement();
        } else {
            m_r->readNextStartElement();
            assertCurrentTag(tag);
            const int n = m_r->attributes().value("n").toInt();
            for (int i = 0; i < n; i++) {
                typename Catalog::CatalogedType* p = nullptr;
                read<Catalog>("E" + QString::number(i), p, argsForConstructor...);
                vec << p;
            }
            gotoEndElementOfTag(tag);
        }
    }

    template <typename Catalog, typename... ArgsForCreation>
    void rw(SelectionProperty<typename Catalog::CatalogedType*>& d,
            ArgsForCreation... argsForCreation)
    {
        if (m_w) {
            typename Catalog::CatalogedType* p = d.get();
            write<Catalog>(d.persistentTag(), p);
        } else {
            typename Catalog::CatalogedType* p = nullptr;
            read<Catalog>(d.persistentTag(), p, argsForCreation...);
            d.set(p);
        }
    }

    //! serialize item which has only one exact class, and can be present or not
    template <typename ItemClass> void rw(const QString& tag, ItemClass*& p)
    {
        if (m_w) {
            m_w->writeStartElement(tag);
            m_w->writeAttribute("valid", p ? "1" : "0");
            if (p != nullptr)
                p->serialize(*this);
            m_w->writeEndElement();
        } else {
            goToStartElementOfTag(tag);
            const bool valid = m_r->attributes().value("valid").toUInt() > 0;
            p = valid ? new ItemClass() : nullptr;
            if (p != nullptr)
                p->serialize(*this);
            gotoEndElementOfTag(tag);
        }
    }

    //! serialize item which has only one exact class, and can be present or not.
    //!
    //! argsForConstructor will be handed to the constructor when creating ItemClass.
    template <typename ItemClass, typename... ArgsForConstructor>
    void rw(const QString& tag, std::unique_ptr<ItemClass>& up,
            ArgsForConstructor... argsForConstructor)
    {
        if (m_w) {
            m_w->writeStartElement(tag);
            m_w->writeAttribute("valid", up ? "1" : "0");
            if (up)
                up->serialize(*this);
            m_w->writeEndElement();
        } else {
            goToStartElementOfTag(tag);
            const bool valid = m_r->attributes().value("valid").toUInt() > 0;
            if (valid) {
                up.reset(new ItemClass(argsForConstructor...));
                up->serialize(*this);
            } else
                up.reset();
            gotoEndElementOfTag(tag);
        }
    }

    //! serialize item which has only one exact class, and can be present or not.
    //!
    //! The given initializer will be called after creation and before de-serialization
    template <typename ItemClass>
    void rwWithInit(const QString& tag, std::unique_ptr<ItemClass>& up,
                    const std::function<void(ItemClass*)>& initializer)
    {
        if (m_w) {
            m_w->writeStartElement(tag);
            m_w->writeAttribute("valid", up ? "1" : "0");
            if (up)
                up->serialize(*this);
            m_w->writeEndElement();
        } else {
            goToStartElementOfTag(tag);
            const bool valid = m_r->attributes().value("valid").toUInt() > 0;
            if (valid) {
                up.reset(new ItemClass());
                if (initializer)
                    initializer(up.get());
                up->serialize(*this);
            } else
                up.reset();
            gotoEndElementOfTag(tag);
        }
    }

    template <typename Catalog> void write(const QString& tag, typename Catalog::CatalogedType* p)
    {
        ASSERT(m_w);
        m_w->writeStartElement(tag);
        m_w->writeAttribute("type", QString::number(static_cast<uint8_t>(Catalog::type(p))));
        if (p != nullptr)
            p->serialize(*this);
        m_w->writeEndElement();
    }

    template <typename Catalog, typename... Args>
    void read(const QString& tag, typename Catalog::CatalogedType*& p, Args... argsForConstructor)
    {
        ASSERT(m_r);
        goToStartElementOfTag(tag);
        const auto type =
            static_cast<typename Catalog::Type>(m_r->attributes().value("type").toUInt());
        p = Catalog::create(type, argsForConstructor...);
        if (p != nullptr)
            p->serialize(*this);
        gotoEndElementOfTag(tag);
    }

    template <typename BaseClass> void rwBaseClass(const QString& tag, BaseClass* p)
    {
        if (m_w) {
            m_w->writeStartElement(tag);
            p->BaseClass::serialize(*this);
            m_w->writeEndElement();
        } else {
            goToStartElementOfTag(tag);
            p->BaseClass::serialize(*this);
            gotoEndElementOfTag(tag);
        }
    }

    //! Serialize a class (not a pointer) which supports void serialize(Serializer&)
    template <typename T> void rw(const QString& tag, T& t)
    {
        if (m_w) {
            m_w->writeStartElement(tag);
            t.serialize(*this);
            m_w->writeEndElement();
        } else {
            goToStartElementOfTag(tag);
            t.serialize(*this);
            gotoEndElementOfTag(tag);
        }
    }

    //! Only for migration
    // #bamigration Remove once the SessionItem migration is done
    template <typename type> void rw(const QString& tag, SessionItem* property)
    {
        if (m_w) {
            auto v = property->value().value<type>();
            rw(tag, v);
        } else {
            type v;
            rw(tag, v);
            property->setValue(v);
        }
    }

private:
    void assertCurrentTag(const QString& expectedTag) const;
    void assertCurrentToken(QXmlStreamReader::TokenType token) const;
    void gotoEndElementOfTag(const QString& tag);
    void goToStartElementOfTag(const QString& tag);

private:
    QXmlStreamWriter* m_w = nullptr;
    QXmlStreamReader* m_r = nullptr;
};


#endif // BORNAGAIN_GUI_MODEL_XML_SERIALIZER_H
